==> toycars-0.3.0.tar.gz: toycars-0.3.0/src/Wheel.cpp <==
/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 31/12/04.
 */

#include "Wheel.h"
#include "Vehicle.h"
#include "GameConstants.h"

Wheel::Wheel(	Vehicle& inParentVehicle,
				PacejkaCurve& inSlipRatioCurve,
				PacejkaCurve& inSlipAngleCurve,
				Tuple inHinge_vc,
				double inRadius,
				double inTurnRadius,
				double inOrientation_vc,
				double inInertia,
				double inMaxBrakeTorque,
				double inRollingResistance)
    :   mVehicle(inParentVehicle),
		
		mSlipRatioCurve(inSlipRatioCurve),
		mSlipAngleCurve(inSlipAngleCurve),
        
        mHinge_vc(inHinge_vc),
        mWheelRadius(inRadius),
        mJointRadius(inTurnRadius),
        
        mOrientation_vc(inOrientation_vc),
        mNorm(-sin(inOrientation_vc), cos(inOrientation_vc)),
        mLocation_vc(mHinge_vc + (mJointRadius * mNorm)),
        
			// default static load calculation
        mLoad(0.25 * 9.8 * mVehicle.getMass()),
        mInertia(inInertia),
        mInvInertia(1.0 / inInertia),
        
        mBrake(0),
		mMaxBrakeTorque(inMaxBrakeTorque),
        mRollingResistance(inRollingResistance),
        
        mVelocity(Tuple(0,0)),
        mForce(Tuple(0,0)),
        mForce_vc(Tuple(0,0)),
        
        mTractionTorque(0),
        mVehicleTorque(0),
        
        mAngularVelocity(0),
        mAngularAcceleration(0),
        
        mSlipRatio(0),
        mLongtitudinalSlipVelocity(0),
        mSlipAngle(0),
		
		mPeakSlipRatio(inSlipRatioCurve.getPeak()),
		mPeakSlipAngle(inSlipAngleCurve.getPeak())
{
}

Wheel::Wheel(Vehicle& inParentVehicle, const Wheel& inWheel)
    :   mVehicle(inParentVehicle),
		
		mSlipRatioCurve(inWheel.mSlipRatioCurve),
		mSlipAngleCurve(inWheel.mSlipAngleCurve),
        
        mHinge_vc(inWheel.mHinge_vc),
        mWheelRadius(inWheel.mWheelRadius),
        mJointRadius(inWheel.mJointRadius),
        
        mOrientation_vc(inWheel.mOrientation_vc),
        mNorm(inWheel.mNorm),
        mLocation_vc(inWheel.mLocation_vc),
        
			// default static load calculation
        mLoad(inWheel.mLoad),
        mInertia(inWheel.mInertia),
        mInvInertia(inWheel.mInvInertia),
        
        mBrake(0),
		mMaxBrakeTorque(inWheel.mMaxBrakeTorque),
        mRollingResistance(inWheel.mRollingResistance),
        
        mVelocity(Tuple(0,0)),
        mForce(Tuple(0,0)),
        mForce_vc(Tuple(0,0)),
        
        mTractionTorque(0),
        mVehicleTorque(0),
        
        mAngularVelocity(0),
        mAngularAcceleration(0),
        
        mSlipRatio(0),
        mLongtitudinalSlipVelocity(0),
        mSlipAngle(0),
		
		mPeakSlipRatio(inWheel.mPeakSlipRatio),
		mPeakSlipAngle(inWheel.mPeakSlipAngle)
{
}

void Wheel::setOrientation(double inOrientation_vc)
{
    mOrientation_vc = inOrientation_vc;
    mNorm.x = -sin(mOrientation_vc);
    mNorm.y = cos(mOrientation_vc);
    mLocation_vc = mHinge_vc + (mJointRadius * mNorm);
}

void Wheel::compute_force_and_torque()
{
    // compute wheel velocity in wheel coords
    
    mVelocity = mNorm.toLocal(mVehicle.mVelocity + (mVehicle.getAngularVelocity() * mLocation_vc.perp()));
    
    // this trick allows the slip angle and slip ratio to
    // behave smoothly at low speeds by avoiding the singularity
    // for a more smooth function, we could use 0.125*sqr(mVelocity.x) + 2.0 if < 4
    // note: does the cut off I'm using depend on scaling? - check this
    
    double virtual_speed = fabs(mVelocity.x);
    if (virtual_speed < 6.0)
        virtual_speed = 6.0;
    
    // compute slip ratio
    
	mTangentVelocity = mAngularVelocity * mWheelRadius;
    mLongtitudinalSlipVelocity = mVelocity.x - mTangentVelocity;
    mSlipRatio = -mLongtitudinalSlipVelocity / virtual_speed;
    
    // compute slip angle
    
    mSlipAngle = atan(-mVelocity.y / virtual_speed);
    
	// calculate normalised slip vector
	
	mSlipVector.x = mSlipRatio / mPeakSlipRatio;
	mSlipVector.y = mSlipAngle / mPeakSlipAngle;
	
	mTotalSlip = mSlipVector.Length();
	
    // compute force - "combined Pacejka" model a la Gregor Veble
	
	if (mTotalSlip > 0)
	{
		mForce.x = mLoad * mSlipVector.x / mTotalSlip * mSlipRatioCurve(mTotalSlip * mPeakSlipRatio);
		mForce.y = mLoad * mSlipVector.y / mTotalSlip * mSlipAngleCurve(mTotalSlip * mPeakSlipAngle);
    }
	else
	{
		mForce.x = 0.0;
		mForce.y = 0.0;
	}
	
    // transform into vehicle coords
    
    mForce_vc = mNorm.fromLocal(mForce);
    
    // compute torques
    
    mTractionTorque = -mWheelRadius * mForce.x;
    mVehicleTorque = mLocation_vc ^ mForce_vc;
    
	// compute torque friction on wheel
	
	// use sign_smooth to smooth the discontinuity
	mTractionTorque -= sign_smooth(mAngularVelocity, 50.0) * (mRollingResistance + mMaxBrakeTorque*mBrake);
    
    // mTractionTorque is the torque created by the traction force.
    // There may be other torques acting on the wheel if it is
    // connected to the drive train, and hence the total torque
    // may be different. These computations must be done by the
    // vehicle class. So the vehicle class MUST also calculate
    // mAngularAcceleration for each wheel.
    
    // mVehicleTorque is the torque this wheel is exerting on the vehicle.
}

==> toycars-0.3.3.tar.gz: toycars-0.3.3/src/Wheel.cpp <==
/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Henner Zilibowitz on 31/12/04.
 */

#include "Wheel.h"
#include "Vehicle.h"
#include "GameConstants.h"

Wheel::Wheel(	Vehicle& inParentVehicle,
				PacejkaCurve& inSlipRatioCurve,
				PacejkaCurve& inSlipAngleCurve,
				Tuple inHinge_vc,
				double inRadius,
				double inTurnRadius,
				double inOrientation_vc,
				double inInertia,
				double inMaxBrakeTorque,
				double inRollingResistance)
    :   mVehicle(inParentVehicle),
		
		mSlipRatioCurve(inSlipRatioCurve),
		mSlipAngleCurve(inSlipAngleCurve),
        
        mHinge_vc(inHinge_vc),
        mWheelRadius(inRadius),
        mJointRadius(inTurnRadius),
        
        mOrientation_vc(inOrientation_vc),
        mNorm(-sin(inOrientation_vc), cos(inOrientation_vc)),
        mLocation_vc(mHinge_vc + (mJointRadius * mNorm)),
        
			// default static load calculation
        mLoad(0.25 * 9.8 * mVehicle.getMass()),
        mInertia(inInertia),
        mInvInertia(1.0 / inInertia),
        
        mBrake(0),
		mMaxBrakeTorque(inMaxBrakeTorque),
        mRollingResistance(inRollingResistance),
        
        mVelocity(Tuple(0,0)),
        mForce(Tuple(0,0)),
        mForce_vc(Tuple(0,0)),
        
        mTractionTorque(0),
        mVehicleTorque(0),
        
        mAngularVelocity(0),
        mAngularAcceleration(0),
        
        mSlipRatio(0),
        mLongtitudinalSlipVelocity(0),
        mSlipAngle(0),
		
		mPeakSlipRatio(inSlipRatioCurve.getPeak()),
		mPeakSlipAngle(inSlipAngleCurve.getPeak())
{
}

Wheel::Wheel(Vehicle& inParentVehicle, const Wheel& inWheel)
    :   mVehicle(inParentVehicle),
		
		mSlipRatioCurve(inWheel.mSlipRatioCurve),
		mSlipAngleCurve(inWheel.mSlipAngleCurve),
        
        mHinge_vc(inWheel.mHinge_vc),
        mWheelRadius(inWheel.mWheelRadius),
        mJointRadius(inWheel.mJointRadius),
        
        mOrientation_vc(inWheel.mOrientation_vc),
        mNorm(inWheel.mNorm),
        mLocation_vc(inWheel.mLocation_vc),
        
			// default static load calculation
        mLoad(inWheel.mLoad),
        mInertia(inWheel.mInertia),
        mInvInertia(inWheel.mInvInertia),
        
        mBrake(0),
		mMaxBrakeTorque(inWheel.mMaxBrakeTorque),
        mRollingResistance(inWheel.mRollingResistance),
        
        mVelocity(Tuple(0,0)),
        mForce(Tuple(0,0)),
        mForce_vc(Tuple(0,0)),
        
        mTractionTorque(0),
        mVehicleTorque(0),
        
        mAngularVelocity(0),
        mAngularAcceleration(0),
        
        mSlipRatio(0),
        mLongtitudinalSlipVelocity(0),
        mSlipAngle(0),
		
		mPeakSlipRatio(inWheel.mPeakSlipRatio),
		mPeakSlipAngle(inWheel.mPeakSlipAngle)
{
}

void Wheel::setOrientation(double inOrientation_vc)
{
    mOrientation_vc = inOrientation_vc;
    mNorm.x = -sin(mOrientation_vc);
    mNorm.y = cos(mOrientation_vc);
    mLocation_vc = mHinge_vc + (mJointRadius * mNorm);
}

void Wheel::compute_force_and_torque()
{
    // compute wheel velocity in wheel coords
    
    mVelocity = mNorm.toLocal(mVehicle.mVelocity + (mVehicle.getAngularVelocity() * mLocation_vc.perp()));
    
    // this trick allows the slip angle and slip ratio to
    // behave smoothly at low speeds by avoiding the singularity
    // for a more smooth function, we could use 0.125*sqr(mVelocity.x) + 2.0 if < 4
    // note: does the cut off I'm using depend on scaling? - check this
    
    double virtual_speed = fabs(mVelocity.x);
    if (virtual_speed < 6.0)
        virtual_speed = 6.0;
    
    // compute slip ratio
    
	mTangentVelocity = mAngularVelocity * mWheelRadius;
    mLongtitudinalSlipVelocity = mVelocity.x - mTangentVelocity;
    mSlipRatio = -mLongtitudinalSlipVelocity / virtual_speed;
    
    // compute slip angle
    
    mSlipAngle = atan(-mVelocity.y / virtual_speed);
    
	// calculate normalised slip vector
	
	mSlipVector.x = mSlipRatio / mPeakSlipRatio;
	mSlipVector.y = mSlipAngle / mPeakSlipAngle;
	
	mTotalSlip = mSlipVector.Length();
	
    // compute force - "combined Pacejka" model a la Gregor Veble
	
	if (mTotalSlip > 0)
	{
		mForce.x = mLoad * mSlipVector.x / mTotalSlip * mSlipRatioCurve(mTotalSlip * mPeakSlipRatio);
		mForce.y = mLoad * mSlipVector.y / mTotalSlip * mSlipAngleCurve(mTotalSlip * mPeakSlipAngle);
    }
	else
	{
		mForce.x = 0.0;
		mForce.y = 0.0;
	}
	
    // transform into vehicle coords
    
    mForce_vc = mNorm.fromLocal(mForce);
    
    // compute torques
    
    mTractionTorque = -mWheelRadius * mForce.x;
    mVehicleTorque = mLocation_vc ^ mForce_vc;
    
	// compute torque friction on wheel
	
	// use sign_smooth to smooth the discontinuity
	mTractionTorque -= sign_smooth(mAngularVelocity, 50.0) * (mRollingResistance + mMaxBrakeTorque*mBrake);
    
    // mTractionTorque is the torque created by the traction force.
    // There may be other torques acting on the wheel if it is
    // connected to the drive train, and hence the total torque
    // may be different. These computations must be done by the
    // vehicle class. So the vehicle class MUST also calculate
    // mAngularAcceleration for each wheel.
    
    // mVehicleTorque is the torque this wheel is exerting on the vehicle.
}
